@@ -1,13 +1,15 @@
 package RT::Extension::LDAPImport;
 
-our $VERSION = '0.31_03';
+# XXX TODO: For historical reasons, this should become 0.33 after 0.32_xx dev releases are done.
+our $VERSION = '0.32_01';
 
 use warnings;
 use strict;
 use base qw(Class::Accessor);
-__PACKAGE__->mk_accessors(qw(_ldap _group screendebug _dnlist));
+__PACKAGE__->mk_accessors(qw(_ldap _group screendebug _users));
 use Carp;
 use Net::LDAP;
+use Net::LDAP::Util qw(escape_filter_value);
 use Data::Dumper;
 
 =head1 NAME
@@ -145,6 +147,9 @@ If it is a scalar, the value will be looked up in LDAP.
 If it is an arrayref, the values will be concatenated 
 together with a single space.
 
+By default users are created as Unprivileged, but you can change this by
+setting $LDAPCreatePrivileged to 1.
+
 =cut
 
 sub import_users {
@@ -161,16 +166,19 @@ sub import_users {
     my $mapping = $RT::LDAPMapping;
     return unless $self->_check_ldap_mapping( mapping => $mapping );
 
-    $self->_dnlist({});
+    $self->_users({});
 
     my $done = 0; my $count = $results->count;
     while (my $entry = $results->shift_entry) {
-        my $user = $self->_build_object( ldap_entry => $entry, skip => qr/(?i)^CF\./, mapping => $mapping );
-        $user->{Name} ||= $user->{EmailAddress};
+        my $user = $self->_build_user_object( ldap_entry => $entry );
         unless ( $user->{Name} ) {
             $self->_warn("No Name or Emailaddress for user, skipping ".Dumper $user);
             next;
         }
+        if ( $user->{Name} =~ /^[0-9]+$/) {
+            $self->_warn("Skipping user '$user->{Name}', as it is numeric");
+            next;
+        }
         $self->_import_user( user => $user, ldap_entry => $entry, import => $args{import} );
         $done++;
         $self->_debug("Imported $done/$count users");
@@ -189,7 +197,7 @@ sub _import_user {
     my %args = @_;
 
     $self->_debug("Processing user $args{user}{Name}");
-    $self->_dnlist->{lc $args{ldap_entry}->dn} = $args{user}{Name};
+    $self->_cache_user( %args );
 
     $args{user} = $self->create_rt_user( %args );
     return unless $args{user};
@@ -200,6 +208,38 @@ sub _import_user {
     return 1;
 }
 
+=head2 _cache_user ldap_entry => Net::LDAP::Entry, [user => { ... }]
+
+Adds the user to a global cache which is used when importing groups later.
+
+Optionally takes a second argument which is a user data object returned by
+_build_user_object.  If not given, _cache_user will call _build_user_object
+itself.
+
+Returns the user Name.
+
+=cut
+
+sub _cache_user {
+    my $self = shift;
+    my %args = (@_);
+    my $user = $args{user} || $self->_build_user_object( ldap_entry => $args{ldap_entry} );
+
+    my $group_map       = $RT::LDAPGroupMapping           || {};
+    my $member_attr_val = $group_map->{Member_Attr_Value} || 'dn';
+    my $membership_key  = lc $member_attr_val eq 'dn'
+                            ? $args{ldap_entry}->dn
+                            : $args{ldap_entry}->get_value($member_attr_val);
+
+    # Fallback to the DN if the user record doesn't have a value
+    unless (defined $membership_key) {
+        $membership_key = $args{ldap_entry}->dn;
+        $self->_warn("User attribute '$member_attr_val' has no value for '$membership_key'; falling back to DN");
+    }
+
+    return $self->_users->{lc $membership_key} = $user->{Name};
+}
+
 sub _show_user_info {
     my $self = shift;
     my %args = @_;
@@ -246,10 +286,28 @@ sub _check_ldap_mapping {
     return 1;
 }
 
+=head2 _build_user_object
+
+Utility method which wraps _build_object to provide sane defaults for building
+users.  It also tries to ensure a Name exists in the returned object.
+
+=cut
+
+sub _build_user_object {
+    my $self = shift;
+    my $user = $self->_build_object(
+        skip    => qr/(?i)^CF\./,
+        mapping => $RT::LDAPMapping,
+        @_
+    );
+    $user->{Name} ||= $user->{EmailAddress};
+    return $user;
+}
+
 =head2 _build_object
 
 Builds up data from LDAP for importing
-Returns a hash of user data ready for RT::User::Create
+Returns a hash of user or group data ready for RT::User::Create or RT::Group::Create
 
 =cut
 
@@ -281,7 +339,7 @@ sub _build_object {
     return $object;
 }
 
-=head3 _parse_ldap_map
+=head3 _parse_ldap_mapping
 
 Internal helper function for import_user
 If we're passed an arrayref, it will recurse 
@@ -358,7 +416,7 @@ sub create_rt_user {
             return;
         } else {
             if ($args{import}) {
-                my ($val, $msg) = $user_obj->Create( %$user, Privileged => 0 );
+                my ($val, $msg) = $user_obj->Create( %$user, Privileged => $RT::LDAPCreatePrivileged ? 1 : 0 );
 
                 unless ($val) {
                     $self->_error("couldn't create user_obj for $user->{Name}: $msg");
@@ -559,6 +617,10 @@ sub import_groups {
             $self->_warn("No Name for group, skipping ".Dumper $group);
             next;
         }
+        if ( $group->{Name} =~ /^[0-9]+$/) {
+            $self->_warn("Skipping group '$group->{Name}', as it is numeric");
+            next;
+        }
         $self->_import_group( %args, group => $group, ldap_entry => $entry );
         $done++;
         $self->_debug("Imported $done/$count groups");
@@ -600,9 +662,9 @@ sub _import_group {
     my $ldap_entry = $args{ldap_entry};
 
     $self->_debug("Processing group $group->{Name}");
-    my $group_obj = $self->create_rt_group( %args, group => $group );
+    my ($group_obj, $created) = $self->create_rt_group( %args, group => $group );
     return if $args{import} and not $group_obj;
-    $self->add_group_members( %args, name => $group->{Name}, group => $group_obj, ldap_entry => $ldap_entry );
+    $self->add_group_members( %args, name => $group->{Name}, group => $group_obj, ldap_entry => $ldap_entry, new => $created );
     return;
 }
 
@@ -629,10 +691,10 @@ sub create_rt_group {
     my $group_obj = RT::Group->new($RT::SystemUser);
     $group_obj->LoadUserDefinedGroup( $group->{Name} );
 
+    my $created;
     if ($group_obj->Id) {
-        my $message = "Group $group->{Name} already exists as ".$group_obj->Id;
         if ($args{import}) {
-            $self->_debug("$message, updating their data");
+            $self->_debug("Group $group->{Name} already exists as ".$group_obj->Id.", updating their data");
             my @results = $group_obj->Update( ARGSRef => $group, AttributesRef => [keys %$group] );
             $self->_debug(join("\n",@results)||'no change');
         } else {
@@ -651,6 +713,7 @@ sub create_rt_group {
                 $self->_error("couldn't create group_obj for $group->{Name}: $msg");
                 return;
             }
+            $created = $val;
             $self->_debug("Created group for $group->{Name} with id ".$group_obj->Id);
         } else {
             print "Found new group $group->{Name} to create in RT\n";
@@ -662,13 +725,13 @@ sub create_rt_group {
     unless ($group_obj->Id) {
         $self->_error("We couldn't find or create $group->{Name}. This should never happen");
     }
-    return $group_obj;
+    return ($group_obj, $created);
 
 }
 
 =head3 add_group_members
 
-Iterate over the list of DNs in the Member_Attr LDAP entry.
+Iterate over the list of values in the Member_Attr LDAP entry.
 Look up the appropriate username from LDAP.
 Add those users to the group.
 Remove members of the RT Group who are no longer members
@@ -692,35 +755,36 @@ sub add_group_members {
         return;
     }
 
-    my $rt_group_members = {};
-    if ($args{group}) {
+    my %rt_group_members;
+    if ($args{group} and not $args{new}) {
         my $user_members = $group->UserMembersObj( Recursively => 0);
         while ( my $member = $user_members->Next ) {
-            $rt_group_members->{$member->Name}++;
+            $rt_group_members{$member->Name} = $member;
         }
     } elsif (not $args{import}) {
         $self->_debug("No group in RT, would create with members:");
     }
 
-    my $dnlist = $self->_dnlist;
+    my $users = $self->_users;
     foreach my $member (@$members) {
         my $username;
-        if (exists $dnlist->{lc $member}) {
-            next unless $username = $dnlist->{lc $member};
+        if (exists $users->{lc $member}) {
+            next unless $username = $users->{lc $member};
         } else {
             my $ldap_users = $self->_run_search(
-                base   => $member,
-                filter => $RT::LDAPFilter,
+                base   => $RT::LDAPBase,
+                filter => "(&$RT::LDAPFilter($RT::LDAPGroupMapping->{Member_Attr_Value}="
+                            . escape_filter_value($member) . "))"
             );
             unless ( $ldap_users && $ldap_users->count ) {
-                $dnlist->{lc $member} = undef;
+                $users->{lc $member} = undef;
                 $self->_error("No user found for $member who should be a member of $groupname");
                 next;
             }
             my $ldap_user = $ldap_users->shift_entry;
-            $dnlist->{lc $member} = $username = $ldap_user->get_value($RT::LDAPMapping->{Name});
+            $username = $self->_cache_user( ldap_entry => $ldap_user );
         }
-        if ( delete $rt_group_members->{$username} ) {
+        if ( delete $rt_group_members{$username} ) {
             $self->_debug("\t$username\tin RT and LDAP");
             next;
         }
@@ -739,17 +803,11 @@ sub add_group_members {
         }
     }
 
-    for my $username (sort keys %$rt_group_members) {
+    for my $username (sort keys %rt_group_members) {
         $self->_debug("\t$username\tin RT, not in LDAP, removing");
         next unless $args{import};
 
-        my $rt_user = RT::User->new($RT::SystemUser);
-        my ($res,$msg) = $rt_user->Load( $username );
-        unless ($res) {
-            $self->_warn("Unable to load $username: $msg");
-            next;
-        }
-        ($res,$msg) = $group->DeleteMember($rt_user->PrincipalObj->Id);
+        my ($res,$msg) = $group->DeleteMember($rt_group_members{$username}->PrincipalObj->Id);
         unless ($res) {
             $self->_warn("Failed to remove $username to $groupname: $msg");
         }
